// XLanguage.h:
// -------------------------------------------------------------------
// This file is a part of "XTL" available at:
// http://www.codecommunity.com
//
// XTL is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License,
// or (at your option) any later version.
// 
// XTL is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with XTL. If not, see <http://www.gnu.org/licenses/>.
//
// coded by: skulk@codecommunity.com
// modified: 2008-01-22 (skulk)
//////////////////////////////////////////////////////////////////////
#pragma once

#include <share.h>
#include <stdio.h>
#include <windows.h>

#include "XLanguageDefinitions.h"



// #######################################################################################



// ***************************************************************************************
// XLanguage usage:
//
// Create a .h file (e.g. MyLanguageMap.h) where you define the language map e.g.
//
// XLanguageMapBegin(MyLanguageMap)
//  XLanguageMapEntry(IDS_MyString, "MyString default value")
//  ...
// XLanguageMapEnd
//
// Create a .cpp file where you define the compiler passes. It _must_ be in a .cpp file as
//  a plain C compiler does not support dynamic initialization. Furthermore it needs 
//  extern "C" linkage when used from C code as it cannot be resolved by the linker
//  otherwise...
//
// extern "C"
// {
// #include "MyLanguageMap.h"
//
// #define XLanguageMapPass_1
// #include "MyLanguageMap.h"
// 
// #define XLanguageMapPass_2
// #include "MyLanguageMap.h"
// 
// #define XLanguageMapPass_3
// #include "MyLanguageMap.h"
// 
// #define XLanguageMapPass_4
// #include "MyLanguageMap.h"
// }
//
// Initialize language handling with a call to:
//
// "_XLanguageMapInit(XLanguageMapGet(MyLangaugeMap)...);"
//
// or
//
// "_XLanguageMapInitFile(XLanguageMapGet(MyLangaugeMap)...);"
//
// You can now directly access the strings by using the ID (e.g. IDS_MyString) which will
//  contain the localized version if available or the predefined default. Not that before
//  initialization and after shutdown strings contain their default values...
//
// Shutdown language map by calling:
//
// "_XLanguageMapShutdown(XLanguageMapGet(MyLangaugeMap));"
//
// The format of the language file looks like this:
//
// IDS_MyString="MyString localized value"
//
// All lines starting with "#" are treated as comments...
//
// Language files should have the following format: modulename[.oem].lang.str
//  Where:
//  "modulename"	= Name of the binary
//  "oem"			= OEM code (optional)
//  "lang"			= ISO 639 code
//  "str"			= Just our extension of choice ;-)
// ***************************************************************************************



// #######################################################################################



// ***************************************************************************************
// XLanguageNames...
// ***************************************************************************************
enum XLanguageNames
{
	XLanguageName_Abkhazian = 0,
	XLanguageName_Afar,
	XLanguageName_Afrikaans,
	XLanguageName_Albanian,
	XLanguageName_Amharic,
	XLanguageName_Arabic,
	XLanguageName_Armenian,
	XLanguageName_Assamese,
	XLanguageName_Aymara,
	XLanguageName_Azerbaijani,
	XLanguageName_Bashkir,
	XLanguageName_Basque,
	XLanguageName_Bengali,
	XLanguageName_Bhutani,
	XLanguageName_Bihari,
	XLanguageName_Bislama,
	XLanguageName_Breton,
	XLanguageName_Bulgarian,
	XLanguageName_Burmese,
	XLanguageName_Byelorussian,
	XLanguageName_Cambodian,
	XLanguageName_Catalan,
	XLanguageName_Chinese,
	XLanguageName_Corsican,
	XLanguageName_Croatian,
	XLanguageName_Czech,
	XLanguageName_Danish,
	XLanguageName_Dutch,
	XLanguageName_English,
	XLanguageName_Esperanto,
	XLanguageName_Estonian,
	XLanguageName_Faroese,
	XLanguageName_Fiji,
	XLanguageName_Finnish,
	XLanguageName_French,
	XLanguageName_Frisian,
	XLanguageName_Galician,
	XLanguageName_Georgian,
	XLanguageName_German,
	XLanguageName_Greek,
	XLanguageName_Greenlandic,
	XLanguageName_Guarani,
	XLanguageName_Gujarati,
	XLanguageName_Hausa,
	XLanguageName_Hebrew,
	XLanguageName_Hindi,
	XLanguageName_Hungarian,
	XLanguageName_Icelandic,
	XLanguageName_Indonesian,
	XLanguageName_Interlingua,
	XLanguageName_Interlingue,
	XLanguageName_Inuktitut,
	XLanguageName_Inupiak,
	XLanguageName_Irish,
	XLanguageName_Italian,
	XLanguageName_Japanese,
	XLanguageName_Javanese,
	XLanguageName_Kannada,
	XLanguageName_Kashmiri,
	XLanguageName_Kazakh,
	XLanguageName_Kinyarwanda,
	XLanguageName_Kirghiz,
	XLanguageName_Kirundi,
	XLanguageName_Korean,
	XLanguageName_Kurdish,
	XLanguageName_Laothian,
	XLanguageName_Latin,
	XLanguageName_Latvian,
	XLanguageName_Lingala,
	XLanguageName_Lithuanian,
	XLanguageName_Macedonian,
	XLanguageName_Malagasy,
	XLanguageName_Malay,
	XLanguageName_Malayalam,
	XLanguageName_Maltese,
	XLanguageName_Maori,
	XLanguageName_Marathi,
	XLanguageName_Moldavian,
	XLanguageName_Mongolian,
	XLanguageName_Nauru,
	XLanguageName_Nepali,
	XLanguageName_Norwegian,
	XLanguageName_Occitan,
	XLanguageName_Oriya,
	XLanguageName_Oromo,
	XLanguageName_Pashto,
	XLanguageName_Persian,
	XLanguageName_Polish,
	XLanguageName_Portuguese,
	XLanguageName_Punjabi,
	XLanguageName_Quechua,
	XLanguageName_RhaetoRomance,
	XLanguageName_Romanian,
	XLanguageName_Russian,
	XLanguageName_Samoan,
	XLanguageName_Sangho,
	XLanguageName_Sanskrit,
	XLanguageName_Scots,
	XLanguageName_Serbian,
	XLanguageName_SerboCroatian,
	XLanguageName_Sesotho,
	XLanguageName_Setswana,
	XLanguageName_Shona,
	XLanguageName_Sindhi,
	XLanguageName_Sinhalese,
	XLanguageName_Siswati,
	XLanguageName_Slovak,
	XLanguageName_Slovenian,
	XLanguageName_Somali,
	XLanguageName_Spanish,
	XLanguageName_Sundanese,
	XLanguageName_Swahili,
	XLanguageName_Swedish,
	XLanguageName_Tagalog,
	XLanguageName_Tajik,
	XLanguageName_Tamil,
	XLanguageName_Tatar,
	XLanguageName_Telugu,
	XLanguageName_Thai,
	XLanguageName_Tibetan,
	XLanguageName_Tigrinya,
	XLanguageName_Tonga,
	XLanguageName_Tsonga,
	XLanguageName_Turkish,
	XLanguageName_Turkmen,
	XLanguageName_Twi,
	XLanguageName_Uighur,
	XLanguageName_Ukrainian,
	XLanguageName_Urdu,
	XLanguageName_Uzbek,
	XLanguageName_Vietnamese,
	XLanguageName_Volapuk,
	XLanguageName_Welsh,
	XLanguageName_Wolof,
	XLanguageName_Xhosa,
	XLanguageName_Yiddish,
	XLanguageName_Yoruba,
	XLanguageName_Zhuang,
	XLanguageName_Zulu,
	XLanguageName_EnumSize
};



// #######################################################################################



// ***************************************************************************************
// ISO 639 codes...
// ***************************************************************************************
static xccharw* g_pszLanguage[XLanguageName_EnumSize] =
{
	__XW("ab"), // Abkhazian
	__XW("aa"), // Afar
	__XW("af"), // Afrikaans
	__XW("sq"), // Albanian
	__XW("am"), // Amharic
	__XW("ar"), // Arabic
	__XW("hy"), // Armenian
	__XW("as"), // Assamese
	__XW("ay"), // Aymara
	__XW("az"), // Azerbaijani
	__XW("ba"), // Bashkir
	__XW("eu"), // Basque
	__XW("bn"), // Bengali
	__XW("dz"), // Bhutani
	__XW("bh"), // Bihari
	__XW("bi"), // Bislama
	__XW("br"), // Breton
	__XW("bg"), // Bulgarian
	__XW("my"), // Burmese
	__XW("be"), // Byelorussian
	__XW("km"), // Cambodian
	__XW("ca"), // Catalan
	__XW("zh"), // Chinese
	__XW("co"), // Corsican
	__XW("hr"), // Croatian
	__XW("cs"), // Czech
	__XW("da"), // Danish
	__XW("nl"), // Dutch
	__XW("en"), // English
	__XW("eo"), // Esperanto
	__XW("et"), // Estonian
	__XW("fo"), // Faroese
	__XW("fj"), // Fiji
	__XW("fi"), // Finnish
	__XW("fr"), // French
	__XW("fy"), // Frisian
	__XW("gl"), // Galician
	__XW("ka"), // Georgian
	__XW("de"), // German
	__XW("el"), // Greek
	__XW("kl"), // Greenlandic
	__XW("gn"), // Guarani
	__XW("gu"), // Gujarati
	__XW("ha"), // Hausa
	__XW("he"), // Hebrew
	__XW("hi"), // Hindi
	__XW("hu"), // Hungarian
	__XW("is"), // Icelandic
	__XW("id"), // Indonesian
	__XW("ia"), // Interlingua
	__XW("ie"), // Interlingue
	__XW("iu"), // Inuktitut
	__XW("ik"), // Inupiak
	__XW("ga"), // Irish
	__XW("it"), // Italian
	__XW("ja"), // Japanese
	__XW("jw"), // Javanese
	__XW("kn"), // Kannada
	__XW("ks"), // Kashmiri
	__XW("kk"), // Kazakh
	__XW("rw"), // Kinyarwanda
	__XW("ky"), // Kirghiz
	__XW("rn"), // Kirundi
	__XW("ko"), // Korean
	__XW("ku"), // Kurdish
	__XW("lo"), // Laothian
	__XW("la"), // Latin
	__XW("lv"), // Latvian
	__XW("ln"), // Lingala
	__XW("lt"), // Lithuanian
	__XW("mk"), // Macedonian
	__XW("mg"), // Malagasy
	__XW("ms"), // Malay
	__XW("ml"), // Malayalam
	__XW("mt"), // Maltese
	__XW("mi"), // Maori
	__XW("mr"), // Marathi
	__XW("mo"), // Moldavian
	__XW("mn"), // Mongolian
	__XW("na"), // Nauru
	__XW("ne"), // Nepali
	__XW("no"), // Norwegian
	__XW("oc"), // Occitan
	__XW("or"), // Oriya
	__XW("om"), // Oromo
	__XW("ps"), // Pashto
	__XW("fa"), // Persian
	__XW("pl"), // Polish
	__XW("pt"), // Portuguese
	__XW("pa"), // Punjabi
	__XW("qu"), // Quechua
	__XW("rm"), // RhaetoRomance
	__XW("ro"), // Romanian
	__XW("ru"), // Russian
	__XW("sm"), // Samoan
	__XW("sg"), // Sangho
	__XW("sa"), // Sanskrit
	__XW("gd"), // Scots
	__XW("sr"), // Serbian
	__XW("sh"), // SerboCroatian
	__XW("st"), // Sesotho
	__XW("tn"), // Setswana
	__XW("sn"), // Shona
	__XW("sd"), // Sindhi
	__XW("si"), // Sinhalese
	__XW("ss"), // Siswati
	__XW("sk"), // Slovak
	__XW("sl"), // Slovenian
	__XW("so"), // Somali
	__XW("es"), // Spanish
	__XW("su"), // Sundanese
	__XW("sw"), // Swahili
	__XW("sv"), // Swedish
	__XW("tl"), // Tagalog
	__XW("tg"), // Tajik
	__XW("ta"), // Tamil
	__XW("tt"), // Tatar
	__XW("te"), // Telugu
	__XW("th"), // Thai
	__XW("bo"), // Tibetan
	__XW("ti"), // Tigrinya
	__XW("to"), // Tonga
	__XW("ts"), // Tsonga
	__XW("tr"), // Turkish
	__XW("tk"), // Turkmen
	__XW("tw"), // Twi
	__XW("ug"), // Uighur
	__XW("uk"), // Ukrainian
	__XW("ur"), // Urdu
	__XW("uz"), // Uzbek
	__XW("vi"), // Vietnamese
	__XW("vo"), // Volapuk
	__XW("cy"), // Welsh
	__XW("wo"), // Wolof
	__XW("xh"), // Xhosa
	__XW("yi"), // Yiddish
	__XW("yo"), // Yoruba
	__XW("za"), // Zhuang
	__XW("zu"), // Zulu
};



// #######################################################################################



// ***************************************************************************************
// Get the ISO 639 code for a language...
// ***************************************************************************************
static xccharw* __XLanguageGetISOCode(xcuint32 uiID)
{
	return g_pszLanguage[uiID < XLanguageName_EnumSize ? uiID : XLanguageName_English];
}

#define _XLanguageGetISOCode __XLanguageGetISOCode



// #######################################################################################



// ***************************************************************************************
// Get the language for an ISO 639 code...
// ***************************************************************************************
static xint32 __XLanguageGetLanguageCode(xccharw* pszISOCode)
{
	xuint32 ui = 0;

	if (!pszISOCode)
		return -1;

	for (ui = 0; ui < (_XCount(g_pszLanguage, xcharw*)); ui++)
	{
		if (_xstricmpw(pszISOCode, g_pszLanguage[ui]) == 0)
		{
			return ui;
		}
	}

	return -1;
}

#define _XLanguageGetLanguageCode __XLanguageGetLanguageCode



// #######################################################################################



// ***************************************************************************************
// Get the language code for a given language ID...
// ***************************************************************************************
static xint32 __XLanguageGetLanguageCodeByLangID(xccharw wID)
{
	switch (PRIMARYLANGID(wID))
	{
		case LANG_AFRIKAANS:
			return XLanguageName_Afrikaans;
		case LANG_ALBANIAN:
			return XLanguageName_Albanian;
		case LANG_ARABIC:
			return XLanguageName_Arabic;
		case LANG_ARMENIAN: 
			return XLanguageName_Armenian;
		case LANG_ASSAMESE:
			return XLanguageName_Assamese;
		case LANG_AZERI:
			// Unsupported...
			return XLanguageName_English;
		case LANG_BASQUE:   
			return XLanguageName_Basque;
		case LANG_BELARUSIAN:
			return XLanguageName_Byelorussian;
		case LANG_BENGALI:
			return XLanguageName_Bengali;
		case LANG_BULGARIAN:
			return XLanguageName_Bulgarian;
		case LANG_CATALAN:
			return XLanguageName_Catalan;
		case LANG_CHINESE: 
			return XLanguageName_Chinese;
		case LANG_CROATIAN:
			return XLanguageName_Croatian;
		case LANG_CZECH:  
			return XLanguageName_Czech;
		case LANG_DANISH:
			return XLanguageName_Danish;
		case LANG_DIVEHI:  
			// Unsupported...
			return XLanguageName_English;
		case LANG_DUTCH:   
			return XLanguageName_Dutch;
		case LANG_ENGLISH:
			return XLanguageName_English;
		case LANG_ESTONIAN:
			return XLanguageName_Estonian;
		case LANG_FAEROESE:
			return XLanguageName_Faroese;
		case LANG_FARSI:
			// Unsupported...
			return XLanguageName_English;
		case LANG_FINNISH:
			return XLanguageName_Finnish;
		case LANG_FRENCH:  
			return XLanguageName_French;
		case LANG_GALICIAN:
			return XLanguageName_Galician;
		case LANG_GEORGIAN:
			return XLanguageName_Georgian;
		case LANG_GERMAN: 
			return XLanguageName_German;
		case LANG_GREEK:   
			return XLanguageName_Greek;
		case LANG_GUJARATI:
			return XLanguageName_Gujarati;
		case LANG_HEBREW:
			return XLanguageName_Hebrew;
		case LANG_HINDI:
			return XLanguageName_Hindi;
		case LANG_HUNGARIAN:
			return XLanguageName_Hungarian;
		case LANG_ICELANDIC:
			return XLanguageName_Icelandic;
		case LANG_INDONESIAN:
			return XLanguageName_Indonesian;
		case LANG_ITALIAN:
			return XLanguageName_Italian;
		case LANG_JAPANESE:  
			return XLanguageName_Japanese;
		case LANG_KANNADA:
			return XLanguageName_Kannada;
		case LANG_KASHMIRI:  
			return XLanguageName_Kashmiri;
		case LANG_KAZAK:
			return XLanguageName_Kazakh;
		case LANG_KONKANI:
			// Unsupported...
			return XLanguageName_English;
		case LANG_KOREAN:  
			return XLanguageName_Korean;
		case LANG_KYRGYZ:   
			return XLanguageName_Kirghiz;
		case LANG_LATVIAN:   
			return XLanguageName_Latvian;
		case LANG_LITHUANIAN:
			return XLanguageName_Lithuanian;
		case LANG_MACEDONIAN:
			return XLanguageName_Macedonian;
		case LANG_MALAY:
			return XLanguageName_Malay;
		case LANG_MALAYALAM:
			return XLanguageName_Malayalam;
		case LANG_MANIPURI:
			// Unsupported...
			return XLanguageName_English;
		case LANG_MARATHI: 
			return XLanguageName_Marathi;
		case LANG_MONGOLIAN:
			return XLanguageName_Mongolian;
		case LANG_NEPALI:
			return XLanguageName_Nepali;
		case LANG_NORWEGIAN:
			return XLanguageName_Norwegian;
		case LANG_ORIYA:
			return XLanguageName_Oriya;
		case LANG_POLISH:
			return XLanguageName_Polish;
		case LANG_PORTUGUESE:
			return XLanguageName_Portuguese;
		case LANG_PUNJABI:
			return XLanguageName_Punjabi;
		case LANG_ROMANIAN:
			return XLanguageName_Romanian;
		case LANG_RUSSIAN:
			return XLanguageName_Russian;
		case LANG_SANSKRIT:
			return XLanguageName_Sanskrit;
		// "LANG_SERBIAN" hat the same value as "LANG_CROATIAN" therefore it's skipped here. I hope no serbian or croatian dude knows this ;-)...
		// case LANG_SERBIAN:
		//	return GetISOCode(Serbian);
		case LANG_SINDHI:
			return XLanguageName_Sindhi;
		case LANG_SLOVAK:
			return XLanguageName_Slovak;
		case LANG_SLOVENIAN:
			return XLanguageName_Slovenian;
		case LANG_SPANISH:
			return XLanguageName_Spanish;
		case LANG_SWAHILI:
			return XLanguageName_Swahili;
		case LANG_SWEDISH:
			return XLanguageName_Swedish;
		case LANG_SYRIAC:
			// Unsupported...
			return XLanguageName_English;
		case LANG_TAMIL:   
			return XLanguageName_Tamil;
		case LANG_TATAR:
			return XLanguageName_Tatar;
		case LANG_TELUGU:
			return XLanguageName_Telugu;
		case LANG_THAI:
			return XLanguageName_Thai;
		case LANG_TURKISH:
			return XLanguageName_Turkish;
		case LANG_UKRAINIAN:
			return XLanguageName_Ukrainian;
		case LANG_URDU:
			return XLanguageName_Urdu;
		case LANG_UZBEK:
			return XLanguageName_Uzbek;
		case LANG_VIETNAMESE:
			return XLanguageName_Vietnamese;

		default:
			// We return an english ISO code by default as a fallback...
			return XLanguageName_English;
	}
};

#define _XLanguageGetLanguageCodeByLangID __XLanguageGetLanguageCodeByLangID



// #######################################################################################



// ***************************************************************************************
// Get the ISO code for a given language ID...
// ***************************************************************************************
static xccharw* __XLanguageGetISOCodeByLangID(xccharw wID)
{
	return _XLanguageGetISOCode(_XLanguageGetLanguageCodeByLangID(wID));
}

#define _XLanguageGetISOCodeByLangID __XLanguageGetISOCodeByLangID



// #######################################################################################



// ***************************************************************************************
// Language file extension...
// ***************************************************************************************
#define _XLanguageExtension __XW(".str")



// #######################################################################################



// ***************************************************************************************
// Buffer size we always use...
// ***************************************************************************************
#define _XLanguageBufferSize 1024



// #######################################################################################



// ***************************************************************************************
// Check if the string is a comment...
// ***************************************************************************************
static xint32 __XLanguageIsComment(xccharw* psz)
{
	if (!psz)
		return 0;

	// Ingore leading spaces/tabs...
	while (psz && (*psz == __XW(' ') || *psz == __XW('\t')))
	{
		psz++;
	}

	return (psz && *psz == __XW('#'));
}

#define _XLanguageIsComment __XLanguageIsComment



// #######################################################################################



// ***************************************************************************************
// Cleanup string...
// ***************************************************************************************
static xint32 __XLanguageCleanupString(xccharw* psz, xcharw* pszResult)
{
	xcharw* pszBegin = NULL;
	xcharw* pszEnd = NULL;

	xuint32 uiSize = 0;

	if (!psz || !pszResult)
		return 0;

	pszBegin = (xcharw*) psz;
	pszEnd = (xcharw*) (psz + (_xstrlenw(psz) - 1));

	// Ingore leading space/tab...
	while (pszBegin && (*pszBegin == __XW(' ') || *pszBegin == __XW('\t')))
	{
		pszBegin++;
	}

	// Ingore trailing space/tab/crlf...
	while (pszEnd && pszEnd != pszBegin && (*pszEnd == __XW(' ') || *pszEnd == __XW('\t') || *pszEnd == __XW('\n') || *pszEnd == __XW('\r')))
	{
		pszEnd--;
	}

	uiSize = (xuint32) (pszEnd - pszBegin + 1) * sizeof(xcharw);

	_XMemoryCopy(pszResult, pszBegin, uiSize);

	return 1;
}

#define _XLanguageCleanupString __XLanguageCleanupString



// #######################################################################################



// ***************************************************************************************
// Parse escape characters...
// ***************************************************************************************
static xint32 __XLanguageParseEscapeCharacters(xccharw* psz, xcharw** pszString)
{
	if (!psz || !pszString)
		return 0;

	xcharw* pszTemp = NULL;
	xcharw szReplacement = __XT('\0');
	xcharw* pszPos = NULL;
	xuint32 uiOffset = 0;

	xuint32 uiSize = _xstrlenw(psz) + 1;

	xcharw* pszSearch = (xcharw*) _xalloc(uiSize * sizeof(xcharw));
	if (!pszSearch)
		return 0;

	_xstrcpyw(pszSearch, uiSize, psz);
	
	while (pszPos = _xstrchrw(pszSearch + uiOffset, __XW('\\')))
	{
		if (!(pszPos + 1))
			break;

		switch (*(pszPos + 1))
		{
			case __XW('n'):
				szReplacement = __XW('\n');
				break;
			case __XW('r'):
				szReplacement = __XW('\r');
				break;
			case __XW('t'):
				szReplacement = __XW('\t');
				break;
			case __XW('\''):
				szReplacement = __XW('\'');
				break;
			case __XW('\"'):
				szReplacement = __XW('\"');
				break;
			case __XW('\\'):
				szReplacement = __XW('\\');
				break;
		}

		if (!szReplacement)
			break;

		uiSize--;

		pszTemp = (xcharw*) _xalloc(uiSize * sizeof(xcharw));
		if (!pszTemp)
		{
			_xfree(pszSearch);

			return 0;
		}

		_XMemoryCopy(pszTemp, pszSearch, (pszPos - pszSearch) * sizeof(xcharw));

		pszTemp[pszPos - pszSearch] = szReplacement;

		_XMemoryCopy(pszTemp + (pszPos - pszSearch + 1), pszSearch + (pszPos - pszSearch + 2), (uiSize - (pszPos - pszSearch) - 1) * sizeof(xcharw));

		_xfree(pszSearch);

		pszSearch = pszTemp;
	}

	*pszString = pszSearch;

	return 1;
}

#define _XLanguageParseEscapeCharacters __XLanguageParseEscapeCharacters



// #######################################################################################



// ***************************************************************************************
// Parse a result...
// ***************************************************************************************
static xint32 __XLanguageParseResult(xccharw* psz, xcharw** pszID, xcharw** pszString)
{
	xcharw* pszSeparator = NULL;
	xcharw* pszEnd = NULL;
	xcharw* pszTemp = NULL;
	xcharw* pszIDTemp = NULL;
	xcharw* pszStringTemp = NULL;
	xuint32 uiSize = 0;
	xint32 iResult = 0;

	if (!psz || !pszID || !pszString)
		return 0;

	pszSeparator = (xcharw*) _xstrchrw(psz, __XW('='));

	pszEnd = (xcharw*) _xstrrchrw(psz, __XW('\"'));

	if (!pszSeparator || !pszEnd || *(pszSeparator + 1) != __XW('\"'))
		return 0;

	uiSize = (xuint32) ((pszSeparator - psz) * sizeof(xcharw));

	pszIDTemp = (xcharw*) _xalloc(uiSize + sizeof(xcharw));
	if (!pszIDTemp)
		return 0;

	_XMemoryCopy(pszIDTemp, psz, uiSize);

	uiSize = (xuint32) ((pszEnd - (pszSeparator + 2)) * sizeof(xcharw));

	pszTemp = (xcharw*) _xalloc(uiSize + sizeof(xcharw));
	if (!pszTemp)
	{
		_xfree(pszIDTemp);

		return 0;
	}

	_XMemoryCopy(pszTemp, pszSeparator + 2, uiSize);

	iResult = _XLanguageParseEscapeCharacters(pszTemp, &pszStringTemp);

	_xfree(pszTemp);

	if (!iResult)
	{
		_xfree(pszIDTemp);

		return 0;
	}

	*pszID = pszIDTemp;

	*pszString = pszStringTemp;

	return 1;
}

#define _XLanguageParseResult __XLanguageParseResult



// #######################################################################################



// ***************************************************************************************
// Parse language file...
// ***************************************************************************************
static xint32 __XLanguageParseFile(const XLanguageMap* pMap, xccharw* pszFileName)
{
	FILE* pFile = NULL;

	xcharw szBuffer[_XLanguageBufferSize] = {L'\0'};
	xcuint32 uiSize = _XCountCharW(szBuffer);

	xcharw szResult[_XLanguageBufferSize] = {L'\0'};

	int iResult = 0;

	xcharw* pszID = NULL;
	xcharw* pszString = NULL;

	xuint32 ui = 0;

	xuint32 uiFound = 0;

	if (!pMap || !pMap->pData || !pMap->psz || !pszFileName)
		return 0;

	if (!(pFile = _wfsopen(pszFileName, __XW("rt, ccs=UNICODE"), _SH_DENYWR)))
		return 0;

	while (!feof(pFile))
	{
		_XMemoryZero(szBuffer, uiSize);
		_XMemoryZero(szResult, uiSize);

		if (fgetws(szBuffer, uiSize, pFile))
		{
			if (_XLanguageIsComment(szBuffer))
				continue;

			_XLanguageCleanupString(szBuffer, szResult);

			if (_xstrlenw(szResult) == 0)
				continue;

			if (iResult = _XLanguageParseResult(szResult, &pszID, &pszString))
			{
				// Insert into lookup map...
				for (ui = 0; pMap->pData[ui].pszID; ui++)
				{
					if (_xstrcmpw(pMap->pData[ui].pszID, pszID) == 0)
					{
						*(pMap->psz[ui]) = pszString;

						uiFound = 1;

						break;
					}
				}

				_xfree(pszID);
				pszID = NULL;

				if (!uiFound)
				{
					_xfree(pszString);
					pszString = NULL;
				}
				else
				{
					uiFound = 0;
				}
			}
		}
	}

	fclose(pFile);

	return iResult;
}

#define _XLanguageParseFile __XLanguageParseFile



// #######################################################################################



// ***************************************************************************************
// Check if a language file is available for a given language ID...
// ***************************************************************************************
static xint32 __XLanguageIsAvailable(xccharw* pszFileName, xccharw* pszLanguage, xcharw* pszFileNameLanguage, xcuint32 uiSizeLanguage)
{
	xcharw sz[_MAX_PATH] = {L'\0'};
	xcuint32 uiSize = _XCountCharW(sz);

	if (!pszFileName || !pszLanguage)
		return 0;

	_xstrcpyw(sz, uiSize, pszFileName);
	_xstrcatw(sz, uiSize, __XW("."));
	_xstrcatw(sz, uiSize, pszLanguage);
	_xstrcatw(sz, uiSize, _XLanguageExtension);

	if (GetFileAttributes(sz) != INVALID_FILE_ATTRIBUTES)
	{
		_xstrcpyw(pszFileNameLanguage, uiSizeLanguage, sz);

		return 1;
	}

	return 0;
}

#define _XLanguageIsAvailable __XLanguageIsAvailable



// #######################################################################################



// ***************************************************************************************
// Initialize languages (file name format of the string file is: name.oem.lang.str)...
// ***************************************************************************************
static xint32 __XLanguageInitFile(const XLanguageMap* pMap, xccharw* pszFileName, xccharw* pszLanguage, xccharw* pszOEM)
{
	xcharw sz[_MAX_PATH] = {__XW('\0')};
	xcharw szFileName[_MAX_PATH] = {__XW('\0')};
	xcuint32 uiSize = _XCountCharW(sz);

	if (!pMap || !pMap->pData || !pMap->psz || !pszFileName)
		return 0;

	_xstrcatw(sz, uiSize, pszFileName);

	if (pszOEM)
	{
		_xstrcatw(sz, uiSize, __XW("."));
		_xstrcatw(sz, uiSize, pszOEM);
	}

	// Check provided language...
	if (pszLanguage)
	{
		// If there's a wish we should respect it...
		if (_XLanguageIsAvailable(sz, pszLanguage, szFileName, uiSize))
		{
			return _XLanguageParseFile(pMap, szFileName);
		}
	}
	else
	{
		if (_XLanguageIsAvailable(sz, _XLanguageGetISOCodeByLangID(GetUserDefaultLangID()), szFileName, uiSize) ||
			_XLanguageIsAvailable(sz, _XLanguageGetISOCodeByLangID(GetSystemDefaultLangID()), szFileName, uiSize))
		{
			return _XLanguageParseFile(pMap, szFileName);
		}
	}	

	return 1;
}

#define _XLanguageInitFile __XLanguageInitFile



// #######################################################################################



// ***************************************************************************************
// Initialize languages (file name format of the string file is: name.oem.lang.str)...
// ***************************************************************************************
static xint32 __XLanguageInit(const XLanguageMap* pMap, HMODULE hModule, xccharw* pszLanguage, xccharw* pszOEM)
{
	xcharw sz[_MAX_PATH] = {__XW('\0')};
	xcuint32 uiSize = _XCountCharW(sz);

	xcharw* psz = NULL;

	if (!pMap || !pMap->pData || !pMap->psz)
		return 0;

	if (!GetModuleFileNameW(hModule, sz, uiSize))
		return 0;

	if (psz = _xstrrchrw(sz, __XW('.')))
	{
		*(psz) = __XW('\0');
	}
	
	return _XLanguageInitFile(pMap, sz, pszLanguage, pszOEM);
}

#define _XLanguageInit __XLanguageInit



// #######################################################################################



// ***************************************************************************************
// Shutdown languages...
// ***************************************************************************************
static xvoid __XLanguageShutdown(const XLanguageMap* pMap)
{
	xuint32 ui = 0;

	// Free lookup map...
	if (!pMap || !pMap->pData || !pMap->psz)
		return;

	for (ui = 0; pMap->pData[ui].pszID; ui++)
	{
		if (*(pMap->psz[ui]) != pMap->pData[ui].pszDefault)
		{
			_xfree((xcharw*) (*(pMap->psz[ui])));

			// Switch back to default - might be better to have something valid...
			*(pMap->psz[ui]) = (xcharw*) (pMap->pData[ui].pszDefault);
		}
	}
}

#define _XLanguageShutdown __XLanguageShutdown



// #######################################################################################
